from typing import Optional
import os
import logging

from langchain import PromptTemplate, HuggingFaceHub, ChatOpenAI, LLMChain

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class LLM:
    def __init__(self,
                 openai_api_key: Optional[str] = None,
                 hf_repo_id: Optional[str] = None,
                 hf_api_token: Optional[str] = None,
                 model_kwargs: Optional[dict] = None):

        # Check if keys are in the environment variables
        openai_api_key = openai_api_key or os.getenv('OPENAI_API_KEY')
        hf_api_token = hf_api_token or os.getenv('HUGGINGFACEHUB_API_TOKEN')

        self.openai_api_key = openai_api_key
        self.hf_repo_id = hf_repo_id
        self.hf_api_token = hf_api_token
        self.model_kwargs = model_kwargs if model_kwargs else {}

        # If the HuggingFace API token is provided, set it in environment variables
        if self.hf_api_token:
            os.environ["HUGGINGFACEHUB_API_TOKEN"] = self.hf_api_token

        # Create the LLM object based on the provided keys
        try:
            if self.hf_repo_id and self.hf_api_token:
                self.llm = HuggingFaceHub(repo_id=self.hf_repo_id, model_kwargs=self.model_kwargs)
            elif self.openai_api_key:
                self.llm = ChatOpenAI(api_key=self.openai_api_key, model_kwargs=self.model_kwargs)
            else:
                raise ValueError("Please provide either OpenAI API key or both HuggingFace repository ID and API token.")
        except Exception as e:
            logger.error("Failed to initialize LLM: %s", e)
            raise
        
    def run(self, prompt: str) -> str:
        template = """Question: {question}
        Answer: Let's think step by step."""
        
        try:
            prompt_template = PromptTemplate(template=template, input_variables=["question"])
            llm_chain = LLMChain(prompt=prompt_template, llm=self.llm)
            return llm_chain.run({"question": prompt})
        except Exception as e:
            logger.error("Failed to generate response: %s", e)
            raise

# example
llm_instance = LLM(openai_api_key="your_openai_key")
result = llm_instance.run("Who won the FIFA World Cup in 1998?")
print(result)

# using HuggingFaceHub
llm_instance = LLM(hf_repo_id="google/flan-t5-xl", hf_api_token="your_hf_api_token")
result = llm_instance.run("Who won the FIFA World Cup in 1998?")
print(result)



from swarms.tools.agent_tools import *
from langchain.tools import BaseTool
from typing import Optional, Type

from langchain.callbacks.manager import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from typing import List, Any, Dict, Optional
from langchain.memory.chat_message_histories import FileChatMessageHistory

import logging
from pydantic import BaseModel, Extra
logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(levelname)s - %(message)s')

class WorkerNode:
    """Useful for when you need to spawn an autonomous agent instance as a worker to accomplish complex tasks, it can search the internet or spawn child multi-modality models to process and generate images and text or audio and so on"""

    def __init__(self, llm, tools, vectorstore):
        self.llm = llm
        self.tools = tools
        self.vectorstore = vectorstore
        self.agent = None

    def create_agent(self, ai_name, ai_role, human_in_the_loop, search_kwargs):
        logging.info("Creating agent in WorkerNode")
        self.agent = AutoGPT.from_llm_and_tools(
            ai_name=ai_name,
            ai_role=ai_role,
            tools=self.tools,
            llm=self.llm,
            memory=self.vectorstore.as_retriever(search_kwargs=search_kwargs),
            human_in_the_loop=human_in_the_loop,
            chat_history_memory=FileChatMessageHistory("chat_history.txt"),
        )
        self.agent.chain.verbose = True

    def add_tool(self, tool: Tool):
        self.tools.append(tool)

    def run(self, prompt: str) -> str:
        if not isinstance(prompt, str):
            raise TypeError("Prompt must be a string")
        
        if not prompt:
            raise ValueError("Prompt is empty")

        self.agent.run([f"{prompt}"])
        return "Task completed by WorkerNode"
    

worker_tool = Tool(
    name="WorkerNode AI Agent",
    func=WorkerNode.run,
    description="Useful for when you need to spawn an autonomous agent instance as a worker to accomplish complex tasks, it can search the internet or spawn child multi-modality models to process and generate images and text or audio and so on"
)


class WorkerNodeInitializer:
    def __init__(self, openai_api_key):
        self.openai_api_key = openai_api_key

    def initialize_llm(self, llm_class, temperature=0.5):
        return llm_class(openai_api_key=self.openai_api_key, temperature=temperature)

    def initialize_tools(self, llm_class):
        llm = self.initialize_llm(llm_class)
        web_search = DuckDuckGoSearchRun()
        tools = [
            web_search,
            WriteFileTool(root_dir=ROOT_DIR),
            ReadFileTool(root_dir=ROOT_DIR),
            process_csv,
            WebpageQATool(qa_chain=load_qa_with_sources_chain(llm)),
        ]
        return tools

    def initialize_vectorstore(self):
        embeddings_model = OpenAIEmbeddings(openai_api_key=self.openai_api_key)
        embedding_size = 1536
        index = faiss.IndexFlatL2(embedding_size)
        return FAISS(embeddings_model.embed_query, index, InMemoryDocstore({}), {})

    def create_worker_node(self, llm_class=ChatOpenAI):
        worker_tools = self.initialize_tools(llm_class)
        vectorstore = self.initialize_vectorstore()
        worker_node = WorkerNode(llm=self.initialize_llm(llm_class), tools=worker_tools, vectorstore=vectorstore)
        worker_node.create_agent(ai_name="Swarm Worker AI Assistant", ai_role="Assistant", human_in_the_loop=False, search_kwargs={})
        return worker_node

def worker_node(openai_api_key):
    initializer = WorkerNodeInitializer(openai_api_key)
    worker_node = initializer.create_worker_node()
    return worker_node


